@@ -13,8 +13,11 @@
 # limitations under the License.
 
 # pyright: reportUnsupportedDunderAll=none
+from logging import getLogger as _getLogger
 
-__version__ = "0.0.43"
+_getLogger("jax.experimental.array_serialization.serialization").setLevel(40)
+
+__version__ = "0.0.57"
 
 __all__ = (
     "aparser",
@@ -23,8 +26,11 @@ __all__ = (
     "escale",
     "executor",
     "jaximus",
+    "loggings",
     "mpric",
     "ops",
     "optimizers",
+    "paths",
     "pytree",
+    "serialization",
 )
